import os

from mock import patch, Mock, ANY
from rireki.core.project import Project
from rireki.stores.digital_ocean import DigitalOcean
from rireki.testing.test_case import TestCase
from rireki.utils.file_helpers import touch
from rireki.utils.string_helpers import str_slug
from rireki.utils.time_helpers import now


class TestDigitalOcean(TestCase):

    def setUp(self):
        TestCase.setUp(self)

        self.store = DigitalOcean()
        self.project = Project(self.faker.name(), None, self.store)

        self.store.project = self.project

    @patch('boto3.client')
    def test_get_files(self, client):
        # Prepare
        region = self.faker.word()
        access_key = self.faker.sentence()
        access_secret = self.faker.sentence()
        bucket = self.faker.word()
        path = str_slug(self.faker.word())

        self.store.load_config({
            'region': region,
            'access_key': access_key,
            'access_secret': access_secret,
            'bucket': bucket,
            'path': path,
        })

        mock_client = Mock()
        mock_paginator = Mock()
        client.return_value = mock_client
        mock_client.get_paginator.return_value = mock_paginator
        mock_paginator.paginate.return_value = [
            {
                'Contents': [
                    {'Key': os.path.join(path, str(now()))},
                ],
            },
        ]

        # Execute
        backup = self.store.get_last_backup()

        # Assert
        assert backup is not None

        client.assert_called_once_with(
            's3',
            region_name=region,
            endpoint_url='https://%s.digitaloceanspaces.com' % region,
            aws_access_key_id=access_key,
            aws_secret_access_key=access_secret,
        )

        mock_client.get_paginator.assert_called_once_with('list_objects')
        mock_paginator.paginate.assert_called_once_with(Bucket=bucket, Prefix=path)

    @patch('boto3.client')
    def test_upload_files(self, client):
        # Prepare
        region = self.faker.word()
        access_key = self.faker.sentence()
        access_secret = self.faker.sentence()
        bucket = self.faker.word()
        path = str_slug(self.faker.word())
        tmp_path = os.path.join(self.home_path, '{}-{}'.format(self.project.slug, now()))
        placeholder_file_name = str_slug(self.faker.word())
        placeholder_file_path = os.path.join(tmp_path, placeholder_file_name)

        self.store.load_config({
            'region': region,
            'access_key': access_key,
            'access_secret': access_secret,
            'bucket': bucket,
            'path': path,
        })

        touch(placeholder_file_path)

        mock_client = Mock()
        client.return_value = mock_client

        # Execute
        self.store.create_backup(tmp_path)

        # Assert
        client.assert_called_once_with(
            's3',
            region_name=region,
            endpoint_url='https://%s.digitaloceanspaces.com' % region,
            aws_access_key_id=access_key,
            aws_secret_access_key=access_secret,
        )

        mock_client.upload_file.assert_called_once_with(
            placeholder_file_path,
            bucket,
            ANY,
        )

        assert mock_client.upload_file.call_args[0][2].startswith(path)
        assert mock_client.upload_file.call_args[0][2].endswith(placeholder_file_name)



from rireki.utils.string_helpers import str_slug


class Project(object):

    def __init__(self, name, driver, store):
        self.name = name
        self.driver = driver
        self.store = store

    @property
    def slug(self):
        return str_slug(self.name)

    def has_pending_backups(self):
        last_backup = self.get_last_backup()

        if not last_backup:
            return True

        return self.driver.has_pending_backups(last_backup.time)

    def get_backups(self):
        return self.store.get_backups()

    def get_last_backup(self):
        return self.store.get_last_backup()

    def perform_backup(self):
        self.driver.perform_backup()



import click

from rireki.core.projects_manager import ProjectsManager


@click.command()
@click.argument('project', required=False)
@click.option(
    '--force',
    is_flag=True,
    help='Perform backups regardless of project being up to date or not',
)
def backup(project=None, force=False):
    """Perform pending backups"""

    if project:
        name = project
        project = ProjectsManager.get_project_by_name(name)

        if not project:
            click.echo('Project with name "%s" is not installed!' % name)
            return

        projects = [project]
    else:
        projects = ProjectsManager.get_projects()

    if not projects:
        click.echo('No projects installed!')
        return

    __process_backups(projects, force)


def __process_backups(projects, force):
    for project in projects:
        __process_backup(project, force)

    click.echo('Done!')


def __process_backup(project, force):
    if not force and not project.has_pending_backups():
        click.echo('Project "%s" does not have any pending backups' % project.name)
        return

    click.echo('Backing up %s...' % project.name)

    try:
        project.perform_backup()
    except Exception as e:
        error_message = click.style('Error: %s' % e, fg='red')

        click.echo(error_message, err=True)



import click
import os
import shutil

from rireki.core.driver import Driver
from rireki.utils.time_helpers import now


class Files(Driver):
    NAME = 'files'

    def __init__(self):
        Driver.__init__(self)

        self.paths = []

    def ask_config(self):
        Driver.ask_config(self)

        self.paths = self.__ask_paths()

    def load_config(self, config):
        Driver.load_config(self, config)

        self.paths = config['paths']

    def get_config(self):
        config = Driver.get_config(self)

        config['paths'] = self.paths

        return config

    def _prepare_backup_files(self, path):
        format = self.__get_archive_format()

        with TemporaryBackupFolder(self) as folder:
            shutil.make_archive(os.path.join(path, 'backup'), format, folder.path)

        return os.path.join(path, 'backup.' + format)

    def __ask_paths(self):
        paths = []
        continue_asking = True

        while continue_asking:
            paths.append(self.__ask_path())

            continue_asking = click.confirm('Is there anything else you\'d like to back up?')

        return paths

    def __ask_path(self):
        path = None

        while not path:
            path = click.prompt('Where are the files you want to back up?')

            if os.path.exists(path):
                break

            if not click.confirm('There is nothing there, are you sure that\'s the correct path?'):
                path = None

        return path

    def __get_archive_format(self):
        supported_formats = [format[0] for format in shutil.get_archive_formats()]

        return 'zip' if 'zip' in supported_formats else 'tar'


class TemporaryBackupFolder():

    def __init__(self, driver):
        self.driver = driver

    def __enter__(self):
        self.path = os.path.join(
            '/tmp',
            'rireki-files-backup-{}-{}'.format(self.driver.project.slug, now())
        )

        os.makedirs(self.path)

        for path in self.driver.paths:
            shutil.copytree(path, os.path.join(self.path, os.path.basename(path)))

        return self

    def __exit__(self, type, value, traceback):
        shutil.rmtree(self.path)



from rireki.stores.aws import AmazonWebServices
from rireki.stores.digital_ocean import DigitalOcean
from rireki.stores.local import Local


stores = {
    AmazonWebServices.NAME: AmazonWebServices,
    DigitalOcean.NAME: DigitalOcean,
    Local.NAME: Local,
}



import re


class Backup(object):

    @classmethod
    def is_backup_name(cls, name):
        return re.search('(\\d+)$', name)

    def __init__(self, name):
        self.name = name

    @property
    def time(self):
        matches = re.findall('(\\d+)$', self.name)

        return int(matches[-1]) if matches else 0



import click
import os
import shutil

from rireki.core.configurable import Configurable
from rireki.utils.string_helpers import str_studly
from rireki.utils.time_helpers import now

DEFAULT_FREQUENCIES = {
    'Hourly': 60,
    'Daily': 1440,
    'Weekly': 10080,
    'Monthly': 43200,
    'Custom': None,
}


class Driver(Configurable):

    def __init__(self):
        Configurable.__init__(self)

        self.project = None
        self.frequency = None

    def ask_config(self):
        Configurable.ask_config(self)

        self.frequency = self.__ask_frequency()

    def load_config(self, config):
        Configurable.load_config(self, config)

        self.frequency = config['frequency']

    def get_config(self):
        config = Configurable.get_config(self)

        config['frequency'] = self.frequency

        return config

    def has_pending_backups(self, last_backup_time):
        frequency_in_seconds = self.frequency * 60

        return last_backup_time < now() - frequency_in_seconds

    def perform_backup(self):
        tmp_path = self._create_temporary_folder()

        try:
            files_path = self._prepare_backup_files(tmp_path)

            self.project.store.create_backup(files_path)
        finally:
            self._clean_backup_files(tmp_path)

    def _create_temporary_folder(self):
        path = '/tmp/rireki-{}-{}-{}'.format(self.name, self.project.slug, now())

        os.makedirs(path)

        return path

    def _clean_backup_files(self, path):
        shutil.rmtree(path)

    def _prepare_backup_files(self, path):
        raise Exception('%s driver must implement _prepare_backup_files method' % self.name)

    def __ask_frequency(self):
        frequency = click.prompt(
            'How often should the backups be performed?',
            type=click.Choice(DEFAULT_FREQUENCIES.keys(), case_sensitive=False),
        )

        frequency = DEFAULT_FREQUENCIES[str_studly(frequency)]

        if not frequency:
            frequency = self.__ask_custom_frequency()

        return frequency

    def __ask_custom_frequency(self):
        return click.prompt(
            'Enter your custom frequency (in minutes)',
            type=click.IntRange(min=1)
        )



import os


def touch(path):
    directory_name = os.path.dirname(path)

    if not os.path.exists(directory_name):
        os.makedirs(directory_name)

    open(path, 'w').close()


def file_put_contents(path, contents):
    touch(path)

    with open(path, 'w') as file:
        file.write(contents)


def file_get_contents(path):
    if not os.path.exists(path):
        return None

    with open(path, 'r') as file:
        return file.read()


def file_get_name(path):
    return file_get_parts(path)['name']


def file_get_extension(path):
    return file_get_parts(path)['extension']


def file_get_parts(path):
    parts = {}

    parts['folder_path'] = os.path.dirname(path)

    basename = os.path.basename(path)
    extension_index = basename.find('.')

    if extension_index == -1:
        parts['name'] = basename
        parts['extension'] = ''
    else:
        parts['name'] = basename[:extension_index]
        parts['extension'] = basename[extension_index+1:]

    return parts



import click

from rireki.core.projects_manager import ProjectsManager
from rireki.utils.array_helpers import array_map
from rireki.utils.output import display_table, format_time
from rireki.utils.time_helpers import now


@click.command()
@click.argument('project', required=False)
def status(project=None):
    """Show status of installed projects"""

    if project:
        name = project
        project = ProjectsManager.get_project_by_name(name)

        if not project:
            click.echo('Project with name "%s" is not installed!' % name)
            return

        __display_project_status(project)
    else:
        projects = ProjectsManager.get_projects()

        __display_projects_status(projects)


def __display_project_status(project):
    if project.has_pending_backups():
        click.echo(click.style('Project needs to be backed up', fg='red'))
    else:
        click.echo(
            click.style('Project ', fg='green') +
            click.style(project.name, bold=True, fg='green') +
            click.style(' is up to date', fg='green'),
        )

    backups = project.get_backups()

    click.echo('')
    click.echo('All backups (%s):' % len(backups))

    for backup in backups:
        click.echo('- %s ago' % format_time(now() - backup.time, 'interval'))


def __display_projects_status(projects):
    if not projects:
        click.echo('No projects installed!')
        return

    display_table(
        ('Name', 'Driver', 'Store', 'Status'),
        array_map(__get_project_info, projects),
    )


def __get_project_info(project):
    if project.has_pending_backups():
        status = {
            'text': 'backup-pending',
            'color': 'red',
        }
    else:
        last_backup_interval = now() - project.get_last_backup().time
        status = {
            'text': 'Backed up %s ago' % format_time(last_backup_interval, 'interval'),
            'color': 'green',
        }

    return (
        project.name,
        project.driver.name,
        project.store.name,
        status,
    )



from rireki.stores.aws import AmazonWebServices


class DigitalOcean(AmazonWebServices):
    NAME = 'digital-ocean'

    def _get_service_name(self):
        return 'Digital Ocean'

    def _get_endpoint_url(self):
        return 'https://%s.digitaloceanspaces.com' % self.region



import click

from rireki.core.project import Project
from rireki.core.projects_manager import ProjectsManager
from rireki.drivers.index import drivers
from rireki.stores.index import stores


@click.command()
@click.argument('name')
@click.option('--driver', type=click.Choice(drivers.keys()), help='Backups driver')
@click.option('--store', type=click.Choice(stores.keys()), help='Backups store')
def add(name, driver=None, store=None):
    """Install a new project"""

    if ProjectsManager.project_exists(name):
        click.echo('Project with name "%s" already installed!' % name)
        return

    driver = __resolve_driver(driver)
    store = __resolve_store(store)

    __add_new_project(name, driver, store)


def __resolve_driver(driver):
    if driver:
        return driver

    return click.prompt(
        'Which driver would you like to use to backup this project?',
        type=click.Choice(drivers.keys()),
    )


def __resolve_store(store):
    if store:
        return store

    return click.prompt(
        'Which store would you like to use to persist this project?',
        type=click.Choice(stores.keys()),
    )


def __add_new_project(project_name, driver_name, store_name):
    driver = __create_driver(driver_name)
    store = __create_store(store_name)
    project = __create_project(project_name, driver, store)

    ProjectsManager.install_project(project)

    click.echo('Project "%s" has been installed!' % project.name)


def __create_driver(name):
    driver = drivers[name]()

    driver.ask_config()

    return driver


def __create_store(name):
    store = stores[name]()

    store.ask_config()

    return store


def __create_project(name, driver, store):
    return Project(name, driver, store)



import json
import os

from rireki.testing.cli import Cli
from rireki.testing.test_case import TestCase
from rireki.utils.file_helpers import touch, file_get_contents
from rireki.utils.output import format_time
from rireki.utils.time_helpers import now, set_testing_now


class TestBackup(TestCase):

    def test_without_installed_projects(self):
        # Execute
        result = Cli.run('backup')

        # Assert
        assert result.exit_code == 0
        assert 'No projects installed!' in result.output

    def test_without_pending_backups(self):
        # Prepare
        project = self._create_project(
            store='local',
            store_config={'path': '/tmp/rireki_testing/store'},
        )

        touch('/tmp/rireki_testing/store/%s/backup' % now())

        # Execute
        result = Cli.run('backup')

        # Assert
        assert result.exit_code == 0
        assert ('Project "%s" does not have any pending backups' % project.name) in result.output
        assert 'Done' in result.output

    def test_with_custom_driver(self):
        # Prepare
        time = now()
        command_output = self.faker.sentence()
        store_path = '/tmp/rireki_testing/store'
        project = self._create_project(
            driver='custom',
            driver_config={'command': 'echo "%s"' % command_output},
            store='local',
            store_config={'path': store_path},
        )

        set_testing_now(time)

        # Execute
        result = Cli.run('backup')

        # Assert
        assert result.exit_code == 0
        assert ('Backing up %s...' % project.name) in result.output
        assert 'Done' in result.output

        backup_path = os.path.join(
            store_path,
            '{}-backup-{}-{}'.format(project.slug, format_time(time, 'date'), time),
            'logs.json',
        )
        assert os.path.exists(backup_path)

        logs = json.loads(file_get_contents(backup_path))
        assert command_output in logs.get('stdout')



from setuptools import setup

setup(
    name='rireki',
    version='0.0.3',
    author='Noel De Martin',
    author_email='noeldemartin@gmail.com',
    description='CLI backup tool.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/noeldemartin/rireki',
    license='MIT',
    packages=[
        'rireki',
        'rireki.commands',
        'rireki.core',
        'rireki.drivers',
        'rireki.stores',
        'rireki.utils',
    ],
    install_requires=['click', 'toml', 'boto3'],
    entry_points={
        'console_scripts': ['rireki = rireki.cli:cli'],
    },
)



import os
import toml

from rireki.testing.cli import Cli
from rireki.testing.test_case import TestCase
from rireki.utils.string_helpers import str_slug


class TestAdd(TestCase):

    def test_existing_project(self):
        # Prepare
        project = self._create_project()

        # Execute
        result = Cli.run('add', project.name)

        # Assert
        assert result.exit_code == 0
        assert ('Project with name "%s" already installed!' % project.name) in result.output

    def test_new_project(self):
        # Prepare
        project_name = self.faker.name()

        # Execute
        result = Cli.run(
            'add', project_name,
            '--driver=custom',
            '--store=local',
            input=('daily', '-', '/tmp'),
        )

        # Assert
        assert result.exit_code == 0
        assert ('Project "%s" has been installed!' % project_name) in result.output

        assert os.path.exists('%s/projects/%s.conf' % (self.home_path, project_name))

        config = toml.load('%s/projects/%s.conf' % (self.home_path, project_name))

        assert 'name' in config
        assert config['name'] == project_name

    def test_new_project_with_files_driver(self):
        # Prepare
        project_name = self.faker.name()
        driver_name = 'files'
        driver_frequency_name = 'daily'
        driver_frequency_minutes = 1440
        driver_paths = [
            '/tmp',
            os.path.join('tmp', str_slug(self.faker.word())),
        ]

        # Execute
        result = Cli.run(
            'add', project_name,
            '--driver=' + driver_name,
            '--store=local',
            input=self.__get_new_project_with_files_driver_input(
                driver_frequency_name,
                driver_paths,
            ),
        )

        # Assert
        assert result.exit_code == 0

        config = toml.load('%s/projects/%s.conf' % (self.home_path, project_name))

        assert 'driver' in config
        assert config['driver']['name'] == driver_name
        assert config['driver']['frequency'] == driver_frequency_minutes
        assert config['driver']['paths'] == driver_paths

    def test_new_project_with_custom_driver(self):
        # Prepare
        project_name = self.faker.name()
        driver_name = 'custom'
        driver_frequency_name = 'custom'
        driver_frequency_minutes = 42
        driver_command = self.faker.word()

        # Execute
        result = Cli.run(
            'add', project_name,
            '--driver=' + driver_name,
            '--store=local',
            input=(driver_frequency_name, str(driver_frequency_minutes), driver_command, '/tmp'),
        )

        # Assert
        assert result.exit_code == 0

        config = toml.load('%s/projects/%s.conf' % (self.home_path, project_name))

        assert 'driver' in config
        assert config['driver']['name'] == driver_name
        assert config['driver']['frequency'] == driver_frequency_minutes
        assert config['driver']['command'] == driver_command

    def test_new_project_with_local_store(self):
        # Prepare
        project_name = self.faker.name()
        store_name = 'local'
        store_path = '/tmp'

        # Execute
        result = Cli.run(
            'add', project_name,
            '--driver=custom',
            '--store=' + store_name,
            input=('daily', '-', store_path),
        )

        # Assert
        assert result.exit_code == 0

        config = toml.load('%s/projects/%s.conf' % (self.home_path, project_name))

        assert 'store' in config
        assert config['store']['name'] == store_name
        assert config['store']['path'] == store_path

    def __get_new_project_with_files_driver_input(self, driver_frequency_name, driver_paths):
        input = []

        # Driver frequency
        input.append(driver_frequency_name)

        # Driver paths - First path
        input.append(driver_paths[0])
        # Driver paths - Yes, add more paths
        input.append('y')
        # Driver paths - Second path
        input.append(driver_paths[1])
        # Driver paths - Yes, path is correct even though it doesn't exist
        input.append('y')
        # Driver paths - No, no more paths
        input.append('N')

        # store path
        input.append('/tmp')

        return input
